/************************************************************
*************************************************************
*                                                           *
* All relevant (system) settings are done in configurations.h *
*                                                           *
*************************************************************
************************************************************/


/****************************************************************************
 Include Section
*****************************************************************************/

#include "configurations.h"
#if defined INSTR_CACHE_EN

#include <cplb.h>
#include <def_LPBlackfin.h>
#include "macros.h"


/*****************************************************************************
 Prototypes
******************************************************************************/

.GLOBAL __icplb_init;
.GLOBAL __icplb_handler;


/*****************************************************************************
 Global Variables
******************************************************************************/

.GLOBAL _b4_icplb_data_pointer;
.GLOBAL _b4_icplb_address_pointer;


/*****************************************************************************
 Variables
******************************************************************************/

.SECTION cplb_data;

.ALIGN 4;
.BYTE4 _b4_icplb_data_pointer = 0;
.BYTE4 _b4_icplb_address_pointer = 0;


/*****************************************************************************
 Symbolic constants / definitions
******************************************************************************/

/* number of entries in _b4_icplbs_table */
#if SDRAM_SIZE == 32
    #if defined __ADSPBF548__
    #define     icplbs_table_length     11
    #else
    #define     icplbs_table_length     10
    #endif
#endif

#if SDRAM_SIZE == 64
    #if defined __ADSPBF548__
    #define     icplbs_table_length     19
    #else
    #define     icplbs_table_length     18
    #endif
#endif

#if (SDRAM_SIZE == 128) || defined (LIMIT32kB)
    #undef icplbs_table_length
    #if defined __ADSPBF548__
    #define     icplbs_table_length     35
    #else
    #define     icplbs_table_length     34
    #endif
#endif


/* number of entries in _b4_dcplbs_table is depending on the PAGE_SIZE */
#define         PAGE_SIZE               PAGE_SIZE_4MB

#if defined LIMIT32kB
    #undef      PAGE_SIZE
    #define     PAGE_SIZE               PAGE_SIZE_1KB
#endif


/***************************************************************************************
L1 Instruction Memory Control Register (IMEM_CONTROL)
RESERVED                [0]     1
ENICPLB                 [1]     Instruction CPLB enable
                                0 - CLPBs disabled, minimal address checking only
         CPLB_ENABLE_ICPLBS     1 - CPLBs enabled
IMC                     [2]     Enable/Disable Instruction Cache functionality
                                0 - Upper 16K byte of LI instruction memory configured as SRAM, also invalidates all cache lines if previously configured as cache
                                1 - Upper 16K byte of L1 instruction memory configured as cache
ILOC                [ 6: 3]     0 - Way not locked
                                1 - Way locked
                                0000 - All ways not locked
                                1111 - All Ways locked ( ILOC0 | ILOC1 | ILOC2 | ILOC3 )
RESERVED            [12: 7]
LRUPRIORST             [13]     0 - LRU priority functionality is enabled
                                1 - All cached LRU priority bits (LRUPRIO) are cleared
RESERVED            [31:14]
***************************************************************************************/

#define IMEM_CONTROL_VAL        ( ENICPLB | IMC )


/***************************************************************************************
ICPLB Data Registers (ICPLB_DATAx)
CPLB_VALID              [0]
                                0 - Invalid (disabled) CPLB entry
                 CPLB_VALID     1 - Valid (enabled) CPLB entry
CPLB_LOCK               [1]     Can be used by software in CPLB replacement algorithms
                                0 - Unlocked, CPLB entry can be replaced
                  CPLB_LOCK     1 - Locked, CPLB entry should not be replaced
CPLB_USER_RD            [2]
                                0 - User mode read access generates protection violation exception
               CPLB_USER_RD     1 - User mode read access permitted
RESERVED            [ 7: 3]
CPLB_LRUPRIO            [8]
                                0 - Low importance
              CPLB_PORTPRIO     1 - High importance
RESERVED            [11: 9]
CPLB_L1_CHBL           [12]     Clear this bit whenever L1 memory is configured as SRAM
                                0 - Non-cacheable in L1
                                1 - Cacheable in L1
RESERVED            [15:14]
PAGE_SIZE           [17:16]
              PAGE_SIZE_1KB     00 - 1K byte page size
              PAGE_SIZE_4KB     01 - 4K byte page size
              PAGE_SIZE_1MB     10 - 1M byte page size
              PAGE_SIZE_4MB     11 - 4M byte page size
RESERVED            [31:18]
***************************************************************************************/

/***************************************************************************************
* Bit 9 of the ICPLB_DATAx is shown as Reserved bit however, when ICACHE is used this  *
* bit needs to be set to the same state as bit 12 (CPLB_L1_CHBL) of this register as a *
* workaround to anomaly: 05000258                                                      *
***************************************************************************************/

#define SET_BIT_9                       0x0200


#define ICPLB_DATA_L1_VAL               ( CPLB_VALID | CPLB_LOCK | PAGE_SIZE_4MB )

#if defined __ADSPBF548__
#define ICPLB_DATA_L2_SRAM_VAL          ( CPLB_VALID             | PAGE_SIZE_1MB | CPLB_USER_RD )
#endif

#define ICPLB_DATA_ASYNC_CACHE_VAL      ( CPLB_VALID | CPLB_LOCK | PAGE_SIZE_4MB                | CPLB_L1_CHBL | SET_BIT_9 )
#define ICPLB_DATA_ASYNC_NOCACHE_VAL    ( CPLB_VALID | CPLB_LOCK | PAGE_SIZE_4MB )

#define ICPLB_DATA_SDRAM_CACHE_VAL      ( CPLB_VALID             | PAGE_SIZE     | CPLB_USER_RD | CPLB_L1_CHBL | SET_BIT_9 )
#define ICPLB_DATA_SDRAM_NOCACHE_VAL    ( CPLB_VALID             | PAGE_SIZE     | CPLB_USER_RD )


/*****************************************************************************
 ICPLB Data
******************************************************************************/

.SECTION cplb_data;
.ALIGN 4;


_b4_icplbs_table:

.BYTE4=

//    0xFFC00000, (ICPLB_DATA_L1_VAL),            /* CORE/SYSTEM MMR */
    0xFFA00000, (ICPLB_DATA_L1_VAL),            /* L1 SRAM */

#if defined __ADSPBF548__
    0xFEB00000, (ICPLB_DATA_L2_SRAM_VAL),       /* L2 SRAM */
#endif

    0x20000000, (ICPLB_DATA_ASYNC_NOCACHE_VAL), /* Async Memory Bank 0-3 */


#if ( SDRAM_SIZE == 64 || SDRAM_SIZE == 32 ) && !defined ( LIMIT32kB )

    0x00000000,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x00400000,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x00800000,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x00C00000,(ICPLB_DATA_SDRAM_CACHE_VAL),
    /* --> 16MB */

#endif


#if ( SDRAM_SIZE == 64 || SDRAM_SIZE == 32 ) && !defined ( LIMIT32kB )

    0x01000000,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x01400000,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x01800000,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x01C00000,(ICPLB_DATA_SDRAM_CACHE_VAL),
    /* --> 32MB */

#endif


#if ( SDRAM_SIZE == 64 ) && !defined ( LIMIT32kB )

    0x02000000,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x02400000,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x02800000,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x02C00000,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x03000000,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x03400000,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x03800000,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x03C00000,(ICPLB_DATA_SDRAM_CACHE_VAL),
    /* --> 64MB */

#endif


#if ( SDRAM_SIZE == 128 ) && !defined ( LIMIT32kB )

    0x04000000,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x04400000,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x04800000,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x04C00000,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x05000000,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x05400000,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x05800000,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x05C00000,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x06000000,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x06400000,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x06800000,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x06C00000,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x07000000,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x07400000,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x07800000,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x07C00000,(ICPLB_DATA_SDRAM_CACHE_VAL),
    /* --> 128MB */

#endif


#if defined LIMIT32kB

    0x00000000,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x00000400,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x00000800,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x00000C00,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x00001000,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x00001400,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x00001800,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x00001C00,(ICPLB_DATA_SDRAM_CACHE_VAL),
    /* --> 8kB */

    0x00002000,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x00002400,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x00002800,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x00002C00,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x00003000,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x00003400,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x00003800,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x00003C00,(ICPLB_DATA_SDRAM_CACHE_VAL),
    /* --> 16kB */
    
    0x00004000,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x00004400,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x00004800,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x00004C00,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x00005000,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x00005400,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x00005800,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x00005C00,(ICPLB_DATA_SDRAM_CACHE_VAL),
    /* --> 24kB */
    
    0x00006000,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x00006400,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x00006800,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x00006C00,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x00007000,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x00007400,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x00007800,(ICPLB_DATA_SDRAM_CACHE_VAL),
    0x00007C00,(ICPLB_DATA_SDRAM_CACHE_VAL),
    /* --> 32kB */

#endif


    /* end of section - termination */
    0xFFFFFFFF;

_b4_icplbs_table.end:


/*****************************************************************************
 ICPLB Code
******************************************************************************/

/*****************************************************************************
 Functions
******************************************************************************/

.SECTION cplb_code;


/****************************************************************************
 Function:      __icplb_init
 
 Description:   Initialize L1 Instruction Memory Registers
                according to the _b4_icplbs_table
*****************************************************************************/


__icplb_init:


    link 0;
    [--SP] = ASTAT;
    [--SP] = (R7:5,P5:4);
    PUSH_DAG_REGx(0);
    PUSH_DAG_REGx(1);
    PUSH_DAG_REGx(2);
    IMM32(P5,SYS_MMR_BASE);


    M0 = 0(z); M1 = 0(z); M2 = 0(z);
    L0 = 0(z); L1 = 0(z); L2 = 0(z);
    B0 = 0(z); B1 = 0(z); B2 = 0(z);


    /* RESET IMEM_CONTROL register */
    R7 = 0x1 (z);
    [P5 + lo(IMEM_CONTROL)] = R7;
    csync;

    /* Initialize ICPLBs */
    IMM32(I0,ICPLB_ADDR0);
    IMM32(P4,_b4_icplb_address_pointer); /* save address pointer for CPBL exception handler required for Round Robin Scheduling */
    R6 = I0;
    [P4] = R6;

    IMM32(I1,ICPLB_DATA0);
    IMM32(P4,_b4_icplb_data_pointer); /* save data pointer for CPBL exception handler required for Round Robin Scheduling */
    R6 = I1;
    [P4] = R6;
    
    IMM32(I2,_b4_icplbs_table);
    R5 = -1; /* end point comparison */
    R6 = 15; /* max counter */

    /* read entries from table */
    read_iaddr:
        R7 = [I2++];
        CC = R7 == R5;
        if CC jump idone;
        [I0++] = R7;
    read_idata:
        R7 = [I2++];
        [I1++] = R7;
        R6 = R6 + R5;
        CC = R6 == R5;
        if !CC jump read_iaddr;
    idone:

    /* ENABLE ICPLBs */
    R7 = IMEM_CONTROL_VAL;
    [P5 + lo(IMEM_CONTROL)] = R7;
    ssync;


    icplb_init.done:
    POP_DAG_REGx(2);
    POP_DAG_REGx(1);
    POP_DAG_REGx(0);
    (R7:5,P5:4) = [SP++];
    ASTAT = [SP++];
    unlink;


    rts;
__icplb_init.END:


/****************************************************************************
 Function:          __icplb_handler
 
 Description:       Identify correct DCPLB entry in ICPLB table
                    Replace one (Round Robin Scheduling) with the correct entry
 
 global variables:  _b4_icplb_address_pointer
                    _b4_icplb_data_pointer

 Registers used:    P5: ICPLB_FAULT_ADDR (fixed)
*****************************************************************************/


__icplb_handler:


    link 0;
    [--SP] = ASTAT;
    [--SP] = (R7:2,P5:3);
    PUSH_DAG_REGx(0);
    PUSH_DAG_REGx(1);
    IMM32(P5,SYS_MMR_BASE);


    /**********************
    * Verify a ICPLB miss *
    **********************/

    R7 = SEQSTAT;
    R6 = 0x3F;
    R7 = R7 & R6;
    R6 = 0x2C; /* Instruction access CPLB miss */
    CC = R7 == R6;
    if !CC jump icplb_handler.done;


    /********************************************
    * Load starting point in _b4_dcplbs_table for: *
    * ICPLB_ADDR -> I0                          *
    * ICPLB_DATA -> I1                          *
    ********************************************/

    IMM32(I0,_b4_icplbs_table); /* ICPLB_ADDRx */
    B0 = I0;
    M0 = 4;
    L0 = 8*icplbs_table_length;

    
    /**************************************************
    * Search for the right entry in the _b4_icplbs_table *
    * depending on ICPLB_FAULT_ADDR                   *
    **************************************************/

    R6 = [P5 + lo(ICPLB_FAULT_ADDR)]; /* ICPLB_FAULT_ADDR */
    IMM32(R1,0x00030000); /* for PAGE_SIZE bits */

    icplb_handler.find_ICPLB_entry:
    R5 = [I0++];    /* ICPLB_ADDRx */
    R4 = [I0++];    /* ICPLB_DATAx */
    /* Extract PAGE_SIZE bits */
    R3 = R4 & R1;
    R3 = R4 >> 16;  /* R3 = CPLB Page Size bits */


    /*****************
    * find PAGE_SIZE *
    *****************/

    CC = R3 == 0x3;
    if CC jump MB4 (BP);    /* PAGE_SIZE = 4MB */
    CC = R3 == 0x2;
    if CC jump MB1;         /* PAGE_SIZE = 1MB */
    CC = R3 == 0x1;
    if CC jump kB4;         /* PAGE_SIZE = 4kB */
                            /* else: PAGE_SIZE = 1kB */

    kB1:                    /* 1kB: 0x0400 */
    R2 = 0x400;
    jump icplb_handler.page_size_End;

    kB4:                    /* 4kB: 0x1000 */
    R2 = 0x1000;
    jump icplb_handler.page_size_End;

    MB1:                    /* 1MB: 0x100000 */
    IMM32(R2,0x00100000);
    jump icplb_handler.page_size_End;

    MB4:                    /* 4MB: 0x400000 */
    IMM32(R2,0x00400000);

    icplb_handler.page_size_End:
    R7 = R5 + R2;           /* Store ICPLB_ADDRx end address */


    /***********************************************************
    * Check if actual entry in _b4_icplbs_table is valid          *
    * ICPLB_ADDRx_START <= ICPLB_FAULT_ADDR < ICPLB_ADDRx_END  *
    * R5 <= R6 < R7                                            *
    ***********************************************************/

    CC = R5 <= R6;
    if !CC jump icplb_handler.find_ICPLB_entry (BP);
    CC = R6 < R7;
    if !CC jump icplb_handler.find_ICPLB_entry (BP);


    /**********************************************
    * Load last index point in _b4_icplbs_table for: *
    * _b4_icplb_address_pointer -> I2                 *
    * _b4_icplb_data_pointer -> I3                    *
    * both need to be updated at the end          *
    * initialization in: __icplb_init              *
    **********************************************/

    IMM32(B0,ICPLB_ADDR0); /* 0xFFE0 0200 */
    IMM32(P4,_b4_icplb_address_pointer);
    R7 = [P4];
    I0 = R7;
    M0 = 4;                     /* 4 byte steps */
    L0 = 64;                    /* 16 ICPLB_ADDRx entrties with each 4 bytes */

    IMM32(B1,ICPLB_DATA0); /* 0xFFE0 0100 */
    IMM32(P3,_b4_icplb_data_pointer);
    R7 = [P3];
    I1 = R7;
    M1 = M0;                    /* 4 byte steps */
    L1 = L0;                    /* 16 ICPLB_DATAx entrties with each 4 bytes */


    /****************************************
    * Check if actual ICPLB entry is locked *
    * If yes than try next one              *
    ****************************************/

    icplb_handler.check_ICPLB_LOCK:
    R2 = [I0++];                /* ICPLB_ADDRx */
    NOP;
    R3 = [I1++];                /* ICPLB_DATAx */
    CC = bittst(R3,1);
    if CC jump icplb_handler.check_ICPLB_LOCK;
    

    /* decrement index because the pointer is already at the next entry */
    R2 = [I0--];                /* ICPLB_ADDRx */
    R3 = [I1--];                /* ICPLB_DATAx */


    /**********************
    * Replace ICPLB entry *
    **********************/

    /* DISABLE ICPLBs to be able to update them */
    R7 = [P5 + lo(IMEM_CONTROL)];
    bitclr (R7,1);
    [P5 + lo(IMEM_CONTROL)] = R7;
    ssync;

    /* Update ICPLB entry and store address in _b4_dcplb_address_pointer and _b4_dcplb_data_pointer for Round Robin scheduling */
    [I0++] = R5;
    R7 = I0;
    [P4] = R7;
    [I1++] = R4;
    R7 = I1;
    [P3] = R7;

    /* ENABLE ICPLBs */
    R7 = IMEM_CONTROL_VAL;
    [P5 + lo(IMEM_CONTROL)] = R7;
    ssync;


    IMM32(P5,IMASK);
    R7 = [P5];
    bitclr (R7,bitpos(EVT_IVG14));
    [P5] = R7;


    icplb_handler.done:
    POP_DAG_REGx(1);
    POP_DAG_REGx(0);
    (R7:1,P5:3) = [SP++];
    ASTAT = [SP++];
    unlink;

    rti;
__icplb_handler.end:


#endif /* defined INSTR_CACHE_EN */


/*****************************************************************************
 EOF
******************************************************************************/
