/****************************************************************************
 Include Section
*****************************************************************************/

#include "global.h"
#include "TWI.h"
#include "SPI.h"
#include "UART.h"

/*****************************************************************************
 Symbolic constants / definitions
******************************************************************************/

#define PUSH_BUTTON_1   PB8
#define PUSH_BUTTON_2   PB9
#define PUSH_BUTTON_3   PB10
#define PUSH_BUTTON_4   PB11
#define PUSH_BUTTON_ALL PB8 | PB9 | PB10 | PB11

#define PUSH_BUTTON     PUSH_BUTTON_ALL
//#define PUSH_BUTTON     PB8 | PB9 | PB10 /* PUSH_BUTTON_4 conflicts with HWAIT signal on EZ-KIT Lite */

/*****************************************************************************
 Functions
******************************************************************************/

.SECTION program;

/*****************************************************************************
 Prototypes
******************************************************************************/

.GLOBAL __pint_run;
//.GLOBAL gpio_ports_setup;
//.GLOBAL interrupts_setup;
//.GLOBAL pint_setup;
//.GLOBAL ivg12_handler;
//.GLOBAL ivg12_handler.pint0;

/****************************************************************************
 Function:      __pint_run
 
 Description:   
*****************************************************************************/

__pint_run:

    link 0;

    call gpio_ports_setup;
    call interrupts_setup;
    call pint_setup;

    unlink;

    rts;
__pint_run.end:


/****************************************************************************
 Function:              pint_setup
 Description:           setup up pin interrupt
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        R7,P5
 Global Registers Used: none
 Global Variables Used: none
 C-Callable :           no
*****************************************************************************/

pint_setup:

    link 0;
    [--SP] = ASTAT;
    [--SP] = (R7:7,P5:5);
    IMM32(P5,SYS_MMR_BASE);


    /* assign push button (PB8-PB11) to PINT0 byte 1 */
    IMM32(R7,B1MAP_PBH);
    [P5 + lo(PINT0_ASSIGN)] = R7;


    IMM32(R7,PUSH_BUTTON);


    /* set to raising edge sensitivity */
    [P5 + lo(PINT0_EDGE_SET)] = R7;
    /* set to level sensitivity */
//    [P5 + lo(PINT0_EDGE_CLEAR)] = R7;


    /* with inverter on, falling edges generate interrupt */
    /* set to falling edge sensitivity */
//    [P5 + lo(PINT0_INVERT_SET)] = R7;
    /* set to raising edge sensitivity */
    [P5 + lo(PINT0_INVERT_CLEAR)] = R7;

    /* W1C potential latches due to history */
    [P5 + lo(PINT0_LATCH)] = R7;

    /* unmask interrupts */
    [P5 + lo(PINT0_MASK_SET)] = R7;


    (R7:7,P5:5) = [SP++];
    ASTAT = [SP++];
    unlink;

    rts;
pint_setup.end:



/****************************************************************************
 Function:              gpio_ports_setup
 Description:           setup up general purpose ports
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        R7:6,P5
 Global Registers Used: none
 Global Variables Used: none
 C-Callable :           no
*****************************************************************************/

gpio_ports_setup:

    link 0;
    [--SP] = ASTAT;
    [--SP] = (R7:6,P5:5);
    IMM32(P5,SYS_MMR_BASE);

    /* enable input drivers for EZ-KIT Lite push button 1 (SW13) on Port B */
    /* pin can be also output or input enabled by other functions */
    R7 = w[P5 + lo(PORTB_INEN)] (z);
    R6 = PUSH_BUTTON (z);
    R7 = R7 | R6;
    w[P5 + lo(PORTB_INEN)] = R7;


    (R7:6,P5:5) = [SP++];
    ASTAT = [SP++];
    unlink;

    rts;
gpio_ports_setup.end:


/****************************************************************************
 Function:              interrupts_setup
 Description:           assign peripheral #19 (PINT0) to IVG12
                        enable IVG12
                        GP Interrupt    SIC_IAR value
                        IVG7            0
                        IVG8            1
                        IVG9            2
                        IVG10           3
                        IVG11           4
                        IVG12           5
                        IVG13           6
                        IVG14           7
                        IVG15           8
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        R7:6,P5
 Global Registers Used: none
 Global Variables Used: none
 C-Callable :           no
*****************************************************************************/

interrupts_setup:

    link 0;
    [--SP] = (R7:6,P5:5);
    IMM32(P5,SYS_MMR_BASE);


    /* Assign the PINT0 interrupt to IVG12 */
    R7 = [P5 + lo(SIC_IAR2)];
    IMM32(R6,0xFFFF0FFF);
    R7 = R7 & R6;
    IMM32(R6,PIVG(19,12));
    R7 = R7 | R6;
    [P5 + lo(SIC_IAR2)] = R7;

    /* Unmask interrupt 19 generated by the PINT0 */
    R7 = [P5 + lo(SIC_IMASK0)];
    bitset(R7, bitpos(IRQ_PINT0));
    [P5 + lo(SIC_IMASK0)] = R7;


    IMM32(P5,CORE_MMR_BASE);

    /* Set up the interrupt vector for the PINT0 */
    IMM32(R7,ivg12_handler);
    [P5 + lo(EVT12)] = R7;

    /* Unmask IVG12 interrupt in the IMASK register */
    R7 = [P5 + lo(IMASK)];
    bitset(R7, bitpos(EVT_IVG12));
    [P5 + lo(IMASK)] = R7;


    (R7:6,P5:5) = [SP++];
    unlink;

    rts;
interrupts_setup.end:


/****************************************************************************
 Function:              ivg12_handler
 Description:           generic IVG12 handler
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        R7,P5,ASTAT
 Global Registers Used: none
 Global Variables Used: none
 C-Callable :           no
*****************************************************************************/

ivg12_handler:

    link 0;
    [--SP] = ASTAT;
    [--SP] = (R7:7,P5:5);
    [--SP] = RETI;

    IMM32(P5,SYS_MMR_BASE);
    /* Was it a pint0 interrupt? */
    R7 = [P5 + lo(SIC_ISR0)];
    CC = bittst(R7, bitpos(IRQ_PINT0));
    if !CC jump ivg12_handler.completed;
    call ivg12_handler.pint0;

    ivg12_handler.completed:
    RETI = [SP++];
    (R7:7,P5:5) = [SP++];
    ASTAT = [SP++];
    unlink;

    rti;
ivg12_handler.end:


/****************************************************************************
 Function:              ivg12_handler.pint0
 Description:           specific Pint 0 handler
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        R7,P5,ASTAT
 Global Registers Used: none
 Global Variables Used: none
 C-Callable :           no
*****************************************************************************/

ivg12_handler.pint0:

    link 0;
    [--SP] = ASTAT;
    [--SP] = (R7:6,P5:5);
    IMM32(P5,SYS_MMR_BASE);


    R7 = [P5 + lo(PINT0_IRQ)];

    
    ivg12_handler.pint0.push_button_1:
    CC = bittst(R7,bitpos(PUSH_BUTTON_1));
    if !CC jump ivg12_handler.pint0.push_button_2;

    call __led1_toogle;
    call __spi_hostboot_run;
    call __led1_toogle;

    R6 = [P5 + lo(PINT0_PINSTATE)];
    CC = bittst(R6,bitpos(PUSH_BUTTON_1));
    if CC jump ivg12_handler.pint0.push_button_1;

    /* clear interrupt request early in the ISR */
    R6= PUSH_BUTTON_1 (z);
    [P5 + lo(PINT0_REQUEST)] = R6;
    ivg12_handler.pint0.push_button_1.end:


    ivg12_handler.pint0.push_button_2:
    CC = bittst(R7,bitpos(PUSH_BUTTON_2));
    if !CC jump ivg12_handler.pint0.push_button_3;

    call __led2_toogle;
    call __twi_hostboot_run;
    call __led2_toogle;

    R6 = [P5 + lo(PINT0_PINSTATE)];
    CC = bittst(R6,bitpos(PUSH_BUTTON_2));
    if CC jump ivg12_handler.pint0.push_button_2;

    /* clear interrupt request early in the ISR */
    R6 = PUSH_BUTTON_2 (z);
    [P5 + lo(PINT0_REQUEST)] = R6;
    ivg12_handler.pint0.push_button_2.end:


    ivg12_handler.pint0.push_button_3:
    CC = bittst(R7,bitpos(PUSH_BUTTON_3));
    if !CC jump ivg12_handler.pint0.push_button_4;

    call __led3_toogle; 
    call __uart_hostboot_run;
    call __led3_toogle;

    R6 = [P5 + lo(PINT0_PINSTATE)];
    CC = bittst(R6,bitpos(PUSH_BUTTON_3));
    if CC jump ivg12_handler.pint0.push_button_3;

    /* clear interrupt request early in the ISR */
    R6 = PUSH_BUTTON_3 (z);
    [P5 + lo(PINT0_REQUEST)] = R6;
    ivg12_handler.pint0.push_button_3.end:


    ivg12_handler.pint0.push_button_4:
    CC = bittst(R7,bitpos(PUSH_BUTTON_4));
    if !CC jump ivg12_handler.pint0.more_service;

    call __led4_toogle;
    call __uart_hostboot_dma_run;
    call __led4_toogle;

    R6 = [P5 + lo(PINT0_PINSTATE)];
    CC = bittst(R6,bitpos(PUSH_BUTTON_4));
    if CC jump ivg12_handler.pint0.push_button_4;

    /* clear interrupt request early in the ISR */
    R6 = PUSH_BUTTON_4 (z);
    [P5 + lo(PINT0_REQUEST)] = R6;
    ivg12_handler.pint0.push_button_4.end:


    /* more service code goes to here */
    ivg12_handler.pint0.more_service:


    ivg12_handler.pint0.all_serviced:
    (R7:6,P5:5) = [SP++];
    ASTAT = [SP++];
    unlink;

    rts;
ivg12_handler.pint0.end:

/*****************************************************************************
 EOF
******************************************************************************/
