/****************************************************************************
 Include Section
*****************************************************************************/

#include "global.h"

/*****************************************************************************
 Symbolic constants / definitions
******************************************************************************/

#define UART1RTS    PE9
#define UART1CTS    PE10
#define UART1TX     PH0
#define UART1RX     PH1

/*****************************************************************************
 Global Variables
******************************************************************************/

.EXTERN loader_file;

/*****************************************************************************
 Variables
******************************************************************************/

.SECTION L1_data;
.ALIGN 4;

/*****************************************************************************
 Functions
******************************************************************************/

.SECTION L1_code;

/*****************************************************************************
 Prototypes
******************************************************************************/

.GLOBAL __uart_hostboot_run;
.GLOBAL __uart_hostboot_dma_run;

//.GLOBAL gpio_ports_setup;
//.GLOBAL uart_setup;
//.GLOBAL uart_set_bitrate;
//.GLOBAL uart_autobaud_host;
//.GLOBAL uart_putc;
//.GLOBAL uart_wait4temt;
//.GLOBAL uart_disable;
//.GLOBAL uart_dma_disable;

/****************************************************************************
 Function:              __uart_hostboot_run
 Description:           do the transfer
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        R7:6,P4:3,ASTAT
 Global Registers Used: P5 -> initially loaded with SYS_MMR_BASE address
 C-Callable :           no
*****************************************************************************/

__uart_hostboot_run:

    link 0;
    [--SP] = ASTAT;
    [--SP] = (R7:6,P5:3);
    PUSH_LOOP_CNTx(0);
    IMM32(P5,SYS_MMR_BASE);


    call gpio_ports_setup;

    /* wait for slave to deassert host wait signal */
    #if defined EVAL_HWAIT
        uart_hostboot_run.initial_wait:
        R6 = w[P5 + lo(PORTE)] (z);
        CC = bittst(R6,bitpos(PFHW));
        if CC jump uart_hostboot_run.initial_wait;
    #endif

    call uart_setup;
    call uart_autobaud_host;

    /* This is our pointer to the LDR file */
    IMM32(P4,loader_file);
    IMM32(P3,length(loader_file));

    lsetup(START,END) LC0 = P3;
    START:  R7 = b[P4++](z);
            #if defined EVAL_HWAIT
                HWAIT: R6 = w[P5 + lo(PORTE)] (z);
                CC = bittst(R6,bitpos(PFHW));
                if CC jump HWAIT;
            #endif
    END:    call uart_putc;

    call uart_disable;


    POP_LOOP_CNTx(0);
    (R7:6,P5:3) = [SP++];
    ASTAT = [SP++];
    unlink;

    rts;
__uart_hostboot_run.END:


/****************************************************************************
 Function:              __uart_hostboot_run
 Description:           do the transfer
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        R7,ASTAT
 Global Registers Used: P3 -> loader file current address pointer
                        P4 -> initial: loader file length [byte], than loader file end address
                        P5 -> initially loaded with SYS_MMR_BASE address
 C-Callable :           no
*****************************************************************************/

__uart_hostboot_dma_run:

    link 0;
    [--SP] = ASTAT;
    [--SP] = (R7:7,P5:2);
    IMM32(P5,SYS_MMR_BASE);


    call gpio_ports_setup;

    /* wait for slave to deassert host wait signal */
    #if defined EVAL_HWAIT
        uart_hostboot_dma_run.initial_wait:
        R7 = w[P5 + lo(PORTE)] (z);
        CC = bittst(R7,bitpos(PFHW));
        if CC jump uart_hostboot_dma_run.initial_wait;
    #endif

    call uart_setup;
    call uart_autobaud_host;

    IMM32(P3,loader_file);          /* P3: start/current address pointer*/
    IMM32(P4,length(loader_file));  /* P4: length, required once to calculate end address */
    P4 = P4+P3;                     /* P4: end address */

    uart_hostboot_dma_run.loop:
    call uart_dma_setup;

    uart_hostboot_dma_run.wait:
    R7 = w[P5 + lo(DMA9_IRQ_STATUS)];
    CC = bittst(R7,bitpos(DMA_DONE));
    if !CC jump uart_hostboot_dma_run.wait;

    R7 = ETBEI (z);
    w[P5 + lo(UART1_IER_CLEAR)] = R7;
    R7 = DMA_DONE (z);
    w[P5 + lo(DMA9_IRQ_STATUS)] = R7;

    CC = P3 == P4; /* current address = end address ? */
    if !CC jump uart_hostboot_dma_run.loop;

    call uart_disable;
    call uart_dma_disable;


    (R7:7,P5:2) = [SP++];
    ASTAT = [SP++];
    unlink;

    rts;
__uart_hostboot_dma_run.END:


/****************************************************************************
 Function:              gpio_ports_setup
 Description:           setup up GP ports
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        R7:6,P5
 Global Registers Used: P5 assumed to contain SYS_MMR_BASE address
 C-Callable :           no
*****************************************************************************/

gpio_ports_setup:

    link 0;
    [--SP] = (R7:6);

    R7 = w[P5 + lo(PORTE_FER)];
    R6 = ( UART1RTS | UART1CTS )(z);
    R7 = R7 | R6;
    w[P5 + lo(PORTE_FER)] = R7;

    R7 = w[P5 + lo(PORTH_FER)];
    R6 = ( UART1TX | UART1RX )(z);
    R7 = R7 | R6;
    w[P5 + lo(PORTH_FER)] = R7;

    /* enable UART1RTS and UART1CTS mux mode */
    R7 = [P5 + lo(PORTE_MUX)];
    IMM32(R6,MUX(0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0));
    R7 = R7 | R6;
    [P5 + lo(PORTE_MUX)] = R7;

    /* enable UART1TX and UART1RX mux mode */
    R7 = [P5 + lo(PORTH_MUX)];
    IMM32(R6,MUX(0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0));
    R7 = R7 | R6;
    [P5 + lo(PORTH_MUX)] = R7;

    #if defined EVAL_HWAIT
        /* Set GPIO direction to input */
        R7 = PFHW;
        w[P5 + lo(PORTE_DIR_CLEAR)] = R7;

        /* Enable PF */
        R7 = w[P5 + lo(PORTE_INEN)](z);
        bitset(R7,bitpos(PFHW));
        w[P5 + lo(PORTE_INEN)] = R7;
    #endif

    (R7:6) = [SP++];
    unlink;

    rts;
gpio_ports_setup.end:


/****************************************************************************
 Function:              uart_setup
 Description:           generic UART setup up function
                        no start bit, 8-bit word length, 2 stop bits
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        R7,P5
 Global Registers Used: P5 assumed to contain SYS_MMR_BASE address
 C-Callable :           no
*****************************************************************************/

uart_setup:

    link 0;
    [--SP] = (R7:7);

    call uart_set_bitrate;

    R7 =   nEGLSI   |   /* Enable Global LS Interrupt */
           nEDBO    |   /* Enable Divide-by-One */
           nFFE     |   /* Force Framing Error on Transmit */
           nFPE     |   /* Force Parity Error on Transmit */
           nRPOLC   |   /* IrDA RX Polarity Change */
           nTPOLC   |   /* IrDA TX Polarity Change */
           nIREN    |   /* Enable IrDA Mode */
            UCEN    (z);/* Enable UART Clocks */
    w[P5 + lo(UART1_GCTL)] = R7;

    #if defined HW_FLOW_CTL
    R7 =    ACTS    |   /* Auto CTS operation for TX handshake */
           nARTS    |   /* Auto RTS generation for RX handshake */
           nFCPOL   |   /* Flow Control Pin Polarity */
           nLOOP_ENA|   /* Loopback Mode Enable */
           nRFRT    |   /* Receive FIFO RTS Threshold */
           nRFIT    |   /* Receive FIFO IRQ Threshold */
           nMRTS    |   /* Manual Request to Send */
           nXOFF    (z);/* Transmitter off */
    #else
    R7 =   nACTS    |   /* Auto CTS operation for TX handshake */
           nARTS    |   /* Auto RTS generation for RX handshake */
           nFCPOL   |   /* Flow Control Pin Polarity */
           nLOOP_ENA|   /* Loopback Mode Enable */
           nRFRT    |   /* Receive FIFO RTS Threshold */
           nRFIT    |   /* Receive FIFO IRQ Threshold */
           nMRTS    |   /* Manual Request to Send */
           nXOFF    (z);/* Transmitter off */
    #endif
    w[P5 + lo(UART1_MCR)] = R7;

    R7 =   nSB      |   /* Set Break */
           nSTP     |   /* Stick Parity */
           nEPS     |   /* Even Parity Select */
           nPEN     |   /* Parity Enable */
           nSTB     |   /* Stop Bits */
//            WLS_8   |   /* 8 data bits */
            WLS(8)  (z);/* Word Length Select */
    w[P5 + lo(UART1_LCR)] = R7;

    R7 =   nERFCI   |   /* Enable Receive FIFO Count Interrupt */
           nETFI    |   /* Enable Transmission Finished Interrupt */
           nEDTPTI  |   /* Enable DMA Transmit PIRQ Interrupt */
           nEDSSI   |   /* Enable Modem Status Interrupt */
           nELSI    |   /* Enable Receive Status Interrupt */
           nETBEI   |   /* Enable Transmit Buffer Empty Interrupt */
           nERBFI   (z);/* Enable Receive Buffer Full Interrupt */
    w[P5 + lo(UART1_IER_SET)] = R7;

    (R7:7) = [SP++];
    unlink;

    rts;
uart_setup.end:


/****************************************************************************
 Function:              uart_dma_setup
 Description:           generic UART DMA setup up function
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        R7:6,P2,ASTAT
 Global Registers Used: P5 -> assumed to contain SYS_MMR_BASE address
                        P3 -> loader file current address pointer
                        P4 -> loader file end address
 C-Callable :           no
*****************************************************************************/

uart_dma_setup:

    link 0;
    [--SP] = ASTAT;
    [--SP] = (R7:6);
    [--SP] = P4;

      R7 = 0 (z);
    w[P5 + lo(DMA9_Y_COUNT)]    = R7;
    w[P5 + lo(DMA9_Y_MODIFY)]   = R7;
      R7 = 1 (z);
    w[P5 + lo(DMA9_X_MODIFY)]   = R7;

     [P5 + lo(DMA9_START_ADDR)] = P3;

    R6 = P3;
    R7 = P4;
    R7 = R7-R6;
    R6 = 0xFFFF (z);
    CC = R7<R6;
    if !CC jump uart_dma_setup.full_block;

    uart_dma_setup.no_full_block:
    w[P5 + lo(DMA9_X_COUNT)] = R7;
    P4 = R7;
    jump uart_dma_setup.init;

    uart_dma_setup.full_block:
    w[P5 + lo(DMA9_X_COUNT)] = R6;
    P4 = R6;

    uart_dma_setup.init:
    R7 =    FLOW_STOP   |   /* Next Operation */
            DI_EN       |   /* Data Interrupt Enable */
           nDI_SEL      |   /* Data Interrupt Timing Select */
            SYNC        |   /* Work Unit Transitions */
           nDMA2D       |   /* DMA Mode */
            WDSIZE_8    |   /* Transfer Word Size */
           nWNR         |   /* DMA Direction */
            DMAEN       (z);/* DMA Channel Enable */
    w[P5 + lo(DMA9_CONFIG)] = R7;

    R7 =   nERFCI       |   /* Enable Receive FIFO Count Interrupt */
           nETFI        |   /* Enable Transmission Finished Interrupt */
           nEDTPTI      |   /* Enable DMA Transmit PIRQ Interrupt */
           nEDSSI       |   /* Enable Modem Status Interrupt */
           nELSI        |   /* Enable Receive Status Interrupt */
            ETBEI       |   /* Enable Transmit Buffer Empty Interrupt */
           nERBFI       (z);/* Enable Receive Buffer Full Interrupt */
    w[P5 + lo(UART1_IER_SET)] = R7;

    P3 = P3+P4; /* update current address pointer */

    P4 = [SP++];
    (R7:6) = [SP++];
    ASTAT = [SP++];
    unlink;

    rts;
uart_dma_setup.end:


/****************************************************************************
 Name:                  uart_set_bitrate
 Description:           calculate and set UART Divisor latch registers
                        UART1_DLH & UART1_DLL that fits former Bit Rate
                        and new system clock
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        R7:6,R1:0,ASTAT
 Global Registers Used: P5 assumed to contain SYS_MMR_BASE address
 Global Variables Used: _b4_UART_BIT_RATE
 C-Callable :           no
*****************************************************************************/

uart_set_bitrate:

    link 0;
    [--SP] = ASTAT;
    [--SP] = (R7:0,P5:5);

    call __get_sclk_hz;      /* R0 = current SCLK [Hz] */
    IMM32(R1,UART_BIT_RATE);

    /* calculate new UART Divisor */
    /* Check EDBO bit: divide by 1 or 16 */
    R7 = w[P5 + lo(UART1_GCTL)];
    CC = bittst(R7,bitpos(EDBO));
    if CC jump uart_set_bitrate.skip_enable_divide_by_16;

    R7 = 0x10;
    R1 *= R7;

    uart_set_bitrate.skip_enable_divide_by_16:

    udiv32(); /* Registers passed: Dividend = R0, Divisor = R1, Quotient -> R0 */

    /* Split UART Divisor to UART1_DLH & UART1_DLL */
    R7 = UARTDLH (z);
    R7 = R0 & R7;
    R7 >>= 8;
    w[P5 + lo(UART1_DLH)] = R7;

    R6 = UARTDLL (z);
    R6 = R0 & R6;
    w[P5 + lo(UART1_DLL)] = R6;

    (R7:0,P5:5) = [SP++];
    ASTAT = [SP++];
    unlink;
    rts;

uart_set_bitrate.end:


/****************************************************************************
 Name:                  uart_autobaud_host
 Description:           send out '@' / 0x40 character for slave autobaud detection
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        R7,ASTAT
 Global Registers Used: P5 -> assumed to contain SYS_MMR_BASE address
 Global Variables Used: none
 C-Callable :           no
*****************************************************************************/

uart_autobaud_host:

    link 0;
    [--SP] = ASTAT;
    [--SP] = R7;

    uart_autobaud_host.wait_thre:
    R7 = w[P5 + lo(UART1_LSR)] (z);
    CC = bittst(R7,bitpos(THRE));
    if !CC jump uart_autobaud_host.wait_thre;
    R7 = 0x40 (z);
    w[P5 + lo(UART1_THR)] = R7;

    #if defined EVAL_HWAIT
        uart_autobaud_host.wait_hwait:
        R7 = w[P5 + lo(PORTE)] (z);
        CC = bittst(R7,bitpos(PFHW));
        if CC jump uart_autobaud_host.wait_hwait;
    #endif

    R7 = [SP++];
    ASTAT = [SP++];
    unlink;
    rts;

uart_autobaud_host.end:


/****************************************************************************
 Name:                  uart_putc
 Description:           This function transmits a character by polling
                        THRE bit in the LSR register.
 Input Parameters:      R7 -> holds the character to transmit
 Return Parameters:     none
 Registers Used:        R6,ASTAT
 Global Registers Used: P5 -> assumed to contain SYS_MMR_BASE address
 Global Variables Used: none
 C-Callable :           no
*****************************************************************************/

uart_putc:

    link 0;
    [--SP] = ASTAT;
    [--SP] = R6;

    uart_putc.wait:
    R6 = w[P5 + lo(UART1_LSR)] (z);
    CC = bittst(R6,bitpos(THRE));
    if !CC jump uart_putc.wait;
    w[P5 + lo(UART1_THR)] = R7;

    R6 = [SP++];
    ASTAT = [SP++];
    unlink;
    rts;

uart_putc.end:


/****************************************************************************
 Name:                  uart_wait4temt
 Description:           This function polls the TEMT bit in the LSR register
                        and waits until all data has been shifted out.
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        R7
 Global Registers Used: P5 -> assumed to contain SYS_MMR_BASE address
 Global Variables Used: none
 C-Callable :           no
*****************************************************************************/

uart_wait4temt:

    link 0;
    [--SP] = ASTAT;
    [--SP] = R7;

    uart_wait4temt.wait:
    R7 = w[P5 + lo(UART1_LSR)] (z);
    CC = bittst(R7,bitpos(TEMT));
    if !CC jump uart_wait4temt.wait;

    R7 = [SP++];
    ASTAT = [SP++];
    unlink;
    rts;

uart_wait4temt.end:


/****************************************************************************
 Name:                  uart_disable
 Description:           This function disables the UART clock,
                        but polls the TEMT bit before
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        R7
 Global Registers Used: P5 -> assumed to contain SYS_MMR_BASE address
 Global Variables Used: none
 C-Callable :           no
*****************************************************************************/

uart_disable:

    link 0;
    [--SP] = R7;

    call uart_wait4temt;
    R7 = 0 (z);
    w[P5 + lo(UART1_GCTL)] = R7;

    R7 = [SP++];
    unlink;
    rts;

uart_disable.end:


/****************************************************************************
 Name:                  uart_dma_disable
 Description:           clear all DMA9 registers
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        R7
 Global Registers Used: P5 -> assumed to contain SYS_MMR_BASE address
 Global Variables Used: none
 C-Callable :           no
*****************************************************************************/

uart_dma_disable:

    link 0;
    [--SP] = R7;

     R7 = 0 (z);
    w[P5 + lo(DMA9_CONFIG)]         = R7;
     [P5 + lo(DMA9_START_ADDR)]     = R7;
     [P5 + lo(DMA9_CURR_ADDR)]      = R7;
     [P5 + lo(DMA9_NEXT_DESC_PTR)]  = R7;
     [P5 + lo(DMA9_CURR_DESC_PTR)]  = R7;
    w[P5 + lo(DMA9_X_COUNT)]        = R7;
    w[P5 + lo(DMA9_X_MODIFY)]       = R7;
    w[P5 + lo(DMA9_Y_COUNT)]        = R7;
    w[P5 + lo(DMA9_Y_MODIFY)]       = R7;

    R7 = [SP++];
    unlink;
    rts;

uart_dma_disable.end:


/*****************************************************************************
 EOF
******************************************************************************/
